#ifndef _JOYSTICK_DATA_H__
#define _JOYSTICK_DATA_H__

#include <string>
#include <vector>
#include <iostream>

#include "xml_serdes.h"

using namespace std;

/** ジョイスティックのデータ
 */
class JoystickData{
public:
	JoystickData():
		time(0),
		button(0),
		axis(0)
	{};

	~JoystickData(){};

	double time;            ///< 取得時間(s)
	string id;               ///< 機器のID (URG0など)

	vector<int> button;     ///< ボタンの状態
	vector<float> axis;     ///< axisの状態

	// type名
	static const char TYPE_ID[]; ///< タイプID文字列

	// 送受信メソッド
	/** データをXMLとして書き出す
	 *
	 * @param [out] os 書き出し先ストリーム
	 * @retval 成功時0, 失敗時は負の値を返す
	 */
	int writeToXml(std::ostream& os) const
	{
		return XmlSerDes::writeToXmlSub(os, *this, TYPE_ID);
	}

	/** XMLからデータを読み出し自分自身を更新する
	 *
	 * @param [in] is 読み出し元ストリーム
	 * @retval 成功時0, 失敗時は負の値を返す
	 */
	int readFromXml(std::istream& is)
	{
		return XmlSerDes::readFromXmlSub(is, *this, TYPE_ID);
	}

	/**
	 * \overload int writeToXml(std::string& s) const
	 * @param [out] s 書き出し先文字列
	 */
	int writeToXml(std::string& s) const
	{
		std::ostringstream os(s);
		return writeToXml(os);
	}

	/** XMLからデータを読み出し自分自身を更新する
	 *
	 * \overload int readFromXml(const std::string& s)
	 * @param [in] s 読み出し元文字列
	 */
	int readFromXml(const std::string& s)
	{
		std::istringstream is(s);
		return readFromXml(is);
	}

private:
	// シリアライズメソッド
	friend class boost::serialization::access;
	template <class Archive>
	void serialize(Archive& ar, const unsigned int /* version */){
		using namespace boost::serialization;
		//using namespace XmlSerDes;

		ar & BOOST_SERIALIZATION_NVP(time);
		ar & BOOST_SERIALIZATION_NVP(button);
		ar & BOOST_SERIALIZATION_NVP(axis);
	};
};

std::ostream& operator<<(std::ostream &os, JoystickData &joy);

#endif //_JOYSTICK_DATA_H__
