#ifndef  _VELOCITY_H__
#define  _VELOCITY_H__

#include <string>
#include <vector>
#include <iostream>

#include "xml_serdes.h"

/** 速度
 */
struct Velocity{
public:
	/** コンストラクタ
	 */
	Velocity():
		time(0.0),
		id(""),
		coordinate("self"),
		vel_x(0),
		vel_y(0),
		vel_the(0)
	{};

	/** デストラクタ
	 */
	~Velocity(){};

	double time;                              ///< 取得時間(s)
	std::string id;                           ///< 機器のID (URG0など)
	std::string coordinate;                   ///< 座標系(self or 9)
	float vel_x, vel_y, vel_the; ///< 速度(m/s, rad/s)

	// type名
	static const char TYPE_ID[];

	// 送受信メソッド
	int writeToXml(std::ostream& os) const
	{
		return XmlSerDes::writeToXmlSub(os, *this, TYPE_ID);
	}

	int readFromXml(std::istream& is)
	{
		return XmlSerDes::readFromXmlSub(is, *this, TYPE_ID);
	}

	int writeToXml(std::string& s) const
	{
		std::ostringstream os(s);
		return writeToXml(os);
	}

	int readFromXml(const std::string& s)
	{
		std::istringstream is(s);
		return readFromXml(is);
	}

	Velocity & operator = (Velocity &vel){
		time = vel.time;
		id = vel.id;
		coordinate = vel.coordinate;
		vel_x = vel.vel_x;
		vel_y = vel.vel_y;
		vel_the = vel.vel_the;
		return *this;
	}

private:
	// シリアライズメソッド
	friend class boost::serialization::access;
	BOOST_SERIALIZATION_SPLIT_MEMBER();
	template <class Archive>
	void save(Archive& ar, const unsigned int /* version */) const{
		using namespace boost::serialization;

		ar & BOOST_SERIALIZATION_NVP(time);
		ar & BOOST_SERIALIZATION_NVP(id);
		ar & BOOST_SERIALIZATION_NVP(coordinate);

		ar & BOOST_SERIALIZATION_NVP(vel_x);
		ar & BOOST_SERIALIZATION_NVP(vel_y);
		ar & BOOST_SERIALIZATION_NVP(vel_the);
	};

	template <class Archive>
	void load(Archive& ar, const unsigned int /* version */){
		using namespace boost::serialization;

		ar & BOOST_SERIALIZATION_NVP(time);
		ar & BOOST_SERIALIZATION_NVP(id);
		ar & BOOST_SERIALIZATION_NVP(coordinate);

		ar & BOOST_SERIALIZATION_NVP(vel_x);
		ar & BOOST_SERIALIZATION_NVP(vel_y);
		ar & BOOST_SERIALIZATION_NVP(vel_the);
	};
};

std::ostream& operator<<(std::ostream &os, Velocity &velocity);

#endif //_VELOCITY_H__
