#include "XmlHandler.h"

#include <iostream>

// for XML
#include <iostream>

#include <xercesc/sax2/SAX2XMLReader.hpp>
#include <xercesc/sax2/Attributes.hpp>
#include <xercesc/sax2/XMLReaderFactory.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
using namespace xercesc;

void XmlHandler::startDocument() {
	flag = 0;
	is_find = false;
}

/**
 * タグの開始を検出したら呼び出されるメソッドです。
 */
void XmlHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
                                 const XMLCh* const qname, const Attributes& attrs) {
	char* name=XMLString::transcode(localname);
	if (!strcmp(name, key.c_str())){
		if (attribute == ""){
			flag = 1;
			is_find = true;
		} else {
			string::size_type equal = attribute.find("=");
			if (equal == string::npos){
				XMLCh* attributeName=XMLString::transcode(attribute.c_str());
				result.push_back(XMLString::transcode(attrs.getValue(attributeName)));
				XMLString::release(&attributeName);
				is_find = true;
			} else {
				string attr = attribute.substr(0, equal);
				XMLCh* attributeName=XMLString::transcode(attr.c_str());
				string value = XMLString::transcode(attrs.getValue(attributeName));
				XMLString::release(&attributeName);

				string::size_type first = attribute.find_first_of("\"");
				string::size_type last   = attribute.find_last_of("\"");
				if (first != string::npos){
					if (value == attribute.substr(first+1, last - first - 1).c_str()){
						flag = 1;
						is_find = true;
					}
				}
			}
		}
	}

	XMLString::release(&name);
}

/**
 * タグの終了を検出したら呼び出されるメソッドです。
 */
void XmlHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
                               const XMLCh* const qname) {
  char* name=XMLString::transcode(localname);
  if (!strcmp(name, key.c_str())) flag = 0;
  XMLString::release(&name);
}

/**
 * タグ内の文字列を検出したら呼び出されるメソッドです。
 */
void XmlHandler::characters(const XMLCh* const chars, const unsigned int length) {
  XMLCh* buffer=new XMLCh[XMLString::stringLen(chars)+1];
  XMLString::copyString(buffer, chars);
  XMLString::trim(buffer);
  char* content=XMLString::transcode(buffer);
  delete[] buffer;

  if (flag) result.push_back(content);
  XMLString::release(&content);
}

SAX2XMLReader* parser;

bool InitXmlParser(void)
{
	try {
		XMLPlatformUtils::Initialize();
	} catch(...) {
		cerr << "Xerces-C++の初期化に失敗しました。" << endl;
		return false;
	}

	parser=XMLReaderFactory::createXMLReader();
	return true;
}

void TerminateXmlParser(void)
{
	delete parser;
	XMLPlatformUtils::Terminate();
}

bool XmlParser(string filename, string key, string attribute, vector<string> *result)
{
	try {
		XmlHandler handler;
		handler.key = key;
		handler.attribute = attribute;
		parser->setContentHandler(&handler);
		parser->parse(filename.c_str());
		*result = handler.result;
		return handler.is_find;
	} catch(...) {
		cerr << "ファイルの解析に失敗しました。" << endl;
	}
	return false;
}

bool XmlParserForString(string data, string key, string attribute, vector<string> *result)
{
	try {
		XmlHandler handler;
		handler.key = key;
		handler.attribute = attribute;

		MemBufInputSource source((const XMLByte*)data.c_str(), data.length(), "memory_buffer");
		parser->setContentHandler(&handler);
		parser->parse(source);
		*result = handler.result;
		return handler.is_find;
	} catch(...) {
		cerr << "ファイルの解析に失敗しました。" << endl;
	}
	return false;
}
