#include <sys/time.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <iomanip>
#include <boost/lexical_cast.hpp>

using namespace std;

#include "baseRTC.h"

#include "xml_serdes.h"
#include "xml_serdes_intl.h"
#include "quantized_vector.h"
#include "navigation_data.h"

using namespace baseRTC;

void
serialize_int(ostream& os)
{
	Header hdr;
	hdr.category = "test";
	hdr.name = "test_of_serialization_int";
	hdr.type = "Int";

	hdr.writeToXml(os);

	int n = 12345;
	writeIntToXml(os, n, "test_integer");
}
void
serialize_double(ostream& os)
{
	Header hdr;
	hdr.category = "test";
	hdr.name = "test_of_serialization_double";
	hdr.type = "Double";

	hdr.writeToXml(os);

	double n = 12345.6789;
	writeDoubleToXml(os, n, "test_double");
}
void
serialize_string(ostream& os)
{
	Header hdr;
	hdr.category = "test";
	hdr.name = "test_of_serialization_string";
	hdr.type = "String";
	
	hdr.writeToXml(os);
	
	string s("test of string serialization");
	writeStringToXml(os, s, "test_string");
}
void
serialize_vector_int(ostream& os)
{
	Header hdr;
	hdr.category = "test";
	hdr.name = "test_of_serialization_vector_int";
	hdr.type = "vector_int";
	
	hdr.writeToXml(os);
	
	string s("test of vector<int> serialization");
	
	vector<int> v;
	for (int i = 0; i < 9; ++i)
		v.push_back(i*i);

	writeVectorToXml(os, v, "vector_of_int");
}
void
serialize_vector_double(ostream& os)
{
	Header hdr;
	hdr.category = "test";
	hdr.name = "test_of_serialization_vector_double";
	hdr.type = "vector_double";
	
	hdr.writeToXml(os);
	
	string s("test of vector<double> serialization");
	
	vector<double> v;
	for (int i = 0; i < 9; ++i)
		v.push_back(i*i);

	writeVectorToXml(os, v, "vector_of_double");
}
void
serialize_bytes(ostream& os)
{
	Header hdr;
	hdr.category = "test";
	hdr.name = "test_of_serialization_bytes";
	hdr.type = "byte_array";
	
	hdr.writeToXml(os);

	vector<unsigned char> v;
	for (int i = 0; i < 256; ++i) {
		v.push_back(i);
	}
	writeBytesToXml(os, v, "bytes_array");
}
void
serialize_qvector(ostream& os)
{
	Header hdr;
	hdr.category = "test";
	hdr.name = "test_of_serialization_qvector";
	hdr.type = "quantized_vector";
	
	hdr.writeToXml(os);


	vector<float> fv;
	for (int i = 0; i < 256; ++i) {
		fv.push_back(float(i) / (256 - 1));
	}

//	QuantizedVector qv;
//	qv.setNumberOfValues(16);
//	qv.setRange(0.0, 1.0);
	QuantizedVector qv(16, 0.0, 1.0);

	qv.assign(fv.begin(), fv.end());

	for (int i = 0; i < 256; ++i) {
		cerr << "in:\t" << i << "\t"
		     << fv[i] << "\t" << qv[i] << "\n";
			
	}
	
	qv.writeToXml(os, "QV");
}






void
serialization_test(ostream& os, size_t idx)
{
	void (* const f_tbl[])(ostream& ) = { // うざ。。
		serialize_int,
		serialize_double,
		serialize_string,
		serialize_vector_int,
		serialize_vector_double,
		serialize_bytes,
		serialize_qvector,
	};
	static const size_t tblsize = sizeof(f_tbl) / sizeof(f_tbl[0]);

	if (idx < tblsize)
		return f_tbl[idx](os);
	else {
		cerr << "Bad index: " << idx << "\n"
		     << " must be 0 ... " << tblsize - 1 << "\n";
		return;
	}
}
void
deserialization_test(istream& ist)
{
	Header hdr;
	hdr.readFromXml(ist);
	cout << "category: " << hdr.category << "\n";
	cout << "name:     " << hdr.name << "\n";
	cout << "type:     " << hdr.type << "\n";
	cout << "time:     " << setprecision(6) << fixed << hdr.time << "\n";
	cout << "version:  " << hdr.hdr_version << "\n";

	ist >> ws;
	cout << "next char == " << char(ist.peek()) << "\n";
	
	if (hdr.type == "Int") {
		cout << "Found Int\n";
		int n = -1;
		readIntFromXml(ist, n, "");
		cout << "Integer :"  << n <<"\n";

	} else if (hdr.type == "Double") {
		cout << "Found Double\n";

		double d = -1.0;
		readDoubleFromXml(ist, d, "");
		cout << "Double :"  << d <<"\n";
	} else if (hdr.type == "String") {
		cout << "Found String\n";

		string s;
		readStringFromXml(ist, s, "");
		cout << "String :"  << s <<"\n";
	} else if (hdr.type == "vector_int") {
		cout << "Found vector_int\n";

		vector<int> v;
		readVectorFromXml(ist, v, "");

		cout << "v.size() == " << v.size() << "\n";
		for (int i = 0, imax=v.size(); i < imax; ++i) {
			cout << "v[" << i << "] = " << v[i] << "\n";
		}

	} else if (hdr.type == "vector_double") {
		cout << "Found vector_double\n";

		vector<double> v;
		readVectorFromXml(ist, v, "");
		cout << "v.size() == " << v.size() << "\n";
		for (int i = 0, imax=v.size(); i < imax; ++i) {
			cout << "v[" << i << "] = " << v[i] << "\n";
		}

	} else if (hdr.type == "byte_array") {
		cout << "Found byte_array\n";

		vector<unsigned char> v;
		readBytesFromXml(ist, v, "");
		cout << "v.size() == " << v.size() << "\n";
		for (int i = 0, imax=v.size(); i < imax; ++i) {
			cout << "v[" << i << "] = " << (int)(v[i]) << "\n";
		}
	} else if (hdr.type == "quantized_vector") {
		cout << "Found quantized vector\n";
		QuantizedVector qv;
		qv.readFromXml(ist, "");
		
		cout << "Number of Values: " << qv.numberOfValues() << "\n";
		cout << "Range min: " << qv.rangeMin() << "\n";
		cout << "Range max: " << qv.rangeMax() << "\n";

		vector<float>fv;
		qv.copy_to(back_inserter(fv));

		for (int i = 0, imax=qv.size(); i < imax; ++i) {
			cout << "qv[" << i << "] = " << qv[i] << "\t"
				<< "fv[" << i << "] = " << fv[i] << "\n";
		}
	} else {
		cout << "unknown!!!\n";
	}
	
}
class Test {
public:
	Test(const QuantizedVector& qv) : qv_(qv) { }
private:
	/* ---- decls for serialization ---- */
	friend class boost::serialization::access;
	template <class Archive>
	void
	serialize(Archive& ar, const unsigned int /* version */) const
	{
		using namespace boost::serialization;
		
		ar & make_nvp("TEST_QV", qv_);
	}
	QuantizedVector qv_;
};
void
test_qvector_serialization()
{
	vector<float> fv;

	for (int i = 0; i < 256; ++i)
		fv.push_back(float(i) / 255);

	QuantizedVector qv(32, 0.0, 1.0);

	qv.assign(fv.begin(), fv.end());

	vector<float> fv2;

	qv.copy_to(back_inserter(fv2));

	for (size_t i = 0; i < fv.size(); ++i) {
		cout << fv[i] << "\t" << fv2[i] << "\n";
	}
	Test test(qv);
	XmlSerDes::writeToXmlSub(cout, test, "hoge");
}

void
exit_with_usage(int eval)
{
	cout << "USAGE: serdes [s|d] command"
	     << "    `s' for serialization `d' for deserialization.\n";
	exit(eval);
}
void
serialize_gnav()
{
	using namespace NavData;
	using namespace XmlSerDes;
	RGIS::PntC p;
	p.x = 0.0;
	p.y = 1.0;
	p.z = 2.0;
	writeToXmlSub(cout, p, "point3d");
	cout << "\n";

	vector<RGIS::PntC> vpc;
	vpc.resize(2);
	vpc[0].x = 1.0;
	vpc[0].y = 2.0;
	vpc[0].z = 3.0;
	vpc[1].x = 2.0;
	vpc[1].y = 4.0;
	vpc[1].z = 6.0;
	
	ShortestPath sp(vpc);
	writeToXmlSub(cout, sp, "path");
	cout << "\n";

	RGIS::BoxC b;
	b.minX = 0.0;
	b.minY = 1.0;
	b.minZ = 2.0;
	b.maxX = 3.0;
	b.maxY = 4.0;
	b.maxZ = 5.0;
	writeToXmlSub(cout, b, "box2d");
	cout << "\n";

	RGIS::RotationC r;
	r.roll = 0.0;
	r.pitch = 1.0;
	r.yaw = 2.0;

	writeToXmlSub(cout, r, "rotation");	
	cout << "\n";

	GridInfo gi;
	gi.numCellsW = 10;
	gi.numCellsH = 11;
	gi.numBands = 12;
	gi.valType = 10;
	gi.noDataVal = 1.5;
	gi.useNoData = false;
	gi.valScale = 3.0;
	gi.cellWidth = 0.25;
	gi.cellHeight = 0.35;
	gi.position = p;
	gi.rotation = r;

	writeToXmlSub(cout, gi, "GridInfo");	
	cout << "\n";

	vector<double> vd;
	for (int i = 0; i < 512; ++i) {
		vd.push_back(i);
	}


	PotentialGrid pg;
	pg.info = gi;
	pg.data.setRange(0.0, 512.0);
	pg.data.setNumberOfValues(16);
	pg.data.assign(vd.begin(), vd.end());

	writeToXmlSub(cout, pg, "PotentialGrid");	
	cout << "\n";

}


int main(int argc, char* argv[])
{
#if 1
//	test_qvector();
//	test_qvector_serialization();

	serialize_gnav();
#else
	if (argc != 2 && argc != 3)
		exit_with_usage(1);

	if (argv[1][0] == 's') {
		if (argc != 3) {
			cout << "s requires command(integer)\n";
			exit_with_usage(4);
		}

		int idx = boost::lexical_cast<int>(argv[2]);
		// BUG: error check 無し

		serialization_test(cout, idx);
	} else if (argv[1][0] == 'd') {
		cout << "Deserialization\n";
		cout << "reading from stdin:" << endl;
		deserialization_test(cin);
	} else {
		cout << "argv[1] == [" << argv[1] << "]\n";
		exit_with_usage(3);
	}
#endif

	return 0;
}
/*
 * Local Variables:
 * mode: c++
 * c-basic-offset: 8
 * indent-tabs-mode: t
 * End:
 *
 */
