// -*- C++ -*-
/*!
 * @file  EnvRecognition.cpp * @brief EnvRecognition * $Date$ 
 *
 * $Id$ 
 */
#include "EnvRecognition.h"
#include "EnvRecognitionDialog.h"

#include "baseRTC.h"

#include <iostream>
#include <iomanip>

#include "SerDesIF.h"
#include "GridMap.h"

#include "Fusion.h"

const char EnvRecognition::Category[] = "navigation";
const char EnvRecognition::Name[]     = "env_recognition";

//using namespace RTC;
using namespace std;

// Module specification
// <rtc-template block="module_spec">

/** コンポーネント情報
 */
static const char* envrecognition_spec[] =
  {
    "implementation_id", "EnvRecognition",
    "type_name",         "EnvRecognition",
    "description",       "EnvRecognition",
    "version",           "1.0.0",
    "vendor",            "Chiba Institute of Technology",
    "category",          "Sensor",
    "activity_type",     "SPORADIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.filename", "./EnvRecognition.xml",
    ""
  };
// </rtc-template>

EnvRecognition::
EnvRecognition(RTC::Manager* manager):
    // <rtc-template block="initializer">
  RTC::DataFlowComponentBase(manager),
	m_outOut("out", m_out)
    // </rtc-template>
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers

  // Set OutPort buffer
  registerOutPort("out", m_outOut);

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

	Config     = NULL;
	Result     = NULL;
	SensorData = NULL;
}

EnvRecognition::
~EnvRecognition()
{
	vector<RTC::TimedString *>::iterator it_m_in;
	vector<RTC::InPort<RTC::TimedString> *>::iterator it_m_inIn;

	if(Config != NULL){
		delete Config;
	}

	if(Result != NULL){
		delete Result;
	}

	if(SensorData != NULL){
		delete SensorData;
	}

	it_m_in = m_in.begin();
	while(it_m_in != m_in.end()){
		delete *it_m_in;
		it_m_in++;
	}

	it_m_inIn = m_inIn.begin();
	while(it_m_inIn != m_inIn.end()){
		delete *it_m_inIn;
		it_m_inIn++;
	}
}


RTC::ReturnCode_t EnvRecognition::
onInitialize()
{
	bindParameter( "filename", filename, "./EnvRecognition.xml");

	int ret;

	try{
		Config = new EnvRecognitionConfig;
	}catch(...){
		cerr << "can't allocate memory for config." << endl;
		return RTC::RTC_ERROR;
	}
	ret = Config->LoadFromFile(filename);
	if(ret != 0){
		cerr << "can't load config file \"" << filename << "\"" << endl;
		delete Config;
		Config = NULL;
		return RTC::RTC_ERROR;
	}
#if 0
	cerr << "onInitialize. load config" << endl;
	cerr << *Config << endl;
#endif

	for(int i = 0; i < Config->input_number; i++){
		char buf[10];
		m_in.push_back(new RTC::TimedString);
		sprintf(buf, "in%d", i);
		m_inIn.push_back(new RTC::InPort<RTC::TimedString>(buf, *m_in[i]));
		registerInPort(buf, *m_inIn[i]);
	}

	delete Config;
	Config = NULL;

	fusion.setWorkMap(&work_map);

	return RTC::RTC_OK;
}

pthread_t tid;

RTC::ReturnCode_t EnvRecognition::
onActivated(RTC::UniqueId ec_id)
{
	int ret;

	// 設定ファイルを読み込み
	Config = new EnvRecognitionConfig;
	ret = Config->LoadFromFile(filename);
	if(ret != 0){
		cerr << "can't load config file \"" << filename << "\"." << endl;
		delete Config;
		Config = NULL;
		return RTC::RTC_ERROR;
	}

	// 設定ファイル内のカテゴリ名とコンポーネント名をチェック
	if(Config->category != EnvRecognition::Category){
		cerr << "Category NOT match (" << Config->category << " != " << EnvRecognition::Category << ")." << endl;
		delete Config;
		Config = NULL;
		return RTC::RTC_ERROR;
	}
	if(Config->name != EnvRecognition::Name){
		cerr << "Name NOT match (" << Config->name << " != " << EnvRecognition::Name << ")." << endl;
		delete Config;
		Config = NULL;
		return RTC::RTC_ERROR;
	}


	// 統合の初期化, パラメータ設定
	fusion.init();
	fusion.setMemoryPara(Config->memory_period, Config->memory_decrease);

	// GUI 初期化
	if (Config->gui){
		initialize_window();
		pthread_create(&tid,NULL,windowThread,NULL);
	}

	// データ送信用インタフェイスクラス初期化
	Result = new SerDesIF<GridMap>(EnvRecognition::Category, EnvRecognition::Name, Config->id);

	Result->Data = Config->grid_map;
	Result->Data.map.clear();
	Result->Data.resize();

	work_map = Config->work_map;
	work_map.map.clear();
	work_map.resize();
	for(int i = 0; i < (int)work_map.map.size(); i ++){
		work_map.map[i] = 0.5f;
	}

	// データ受信用インタフェイスクラス初期化
	SensorData = new SerDesIF<GridMap>();
	PosData =    new SerDesIF<Localization>();

	return RTC::RTC_OK;
}

RTC::ReturnCode_t EnvRecognition::
onDeactivated(RTC::UniqueId ec_id)
{
	// GUI終了処理
	if (Config->gui){
		close_window();
		pthread_join(tid, NULL);
	}

	// 統合終了処理
	fusion.close();

	// オドメトリデータ用クラス破棄
	if(PosData != NULL){
		delete PosData;
		PosData = NULL;
	}

	// センサデータ用クラス破棄
	if(SensorData != NULL){
		delete SensorData;
		SensorData = NULL;
	}

	// 結果送信用クラス破棄
	if(Result != NULL){
		delete Result;
		Result = NULL;
	}

	// コンフィグ破棄
	if(Config != NULL){
		delete Config;
		Config = NULL;
	}

	return RTC::RTC_OK;
}

RTC::ReturnCode_t EnvRecognition::
onExecute(RTC::UniqueId ec_id)
{
	baseRTC::Header header;
	string recv;
	vector<string> buf;

	int i;
	bool is_update = false;

	for(i = 0; i < Config->input_number; i++){
		if (m_inIn[i]->isNew()){
			m_inIn[i]->read();
			recv = m_in[i]->data;

#if 0
			cout << "+++++ XML to recv. +++++" << endl;
			cerr << recv << endl;
			cout << "++++++++++++++++++++++++" << endl;
#endif

			header.readFromXml(recv);

			// 受信したデータのIDが, 設定ファイルのreceive 項目にあれば処理
			if(find(Config->receive.begin(), Config->receive.end(), header.id) != Config->receive.end()){
				// ヘッダ内のデータ情報から受信データの型を取得
				if(header.sender_name == "TOP_URG"){
					ProcessURGData(recv);
					Result->Data.time = SensorData->Data.time;
					is_update = true;
				}else if(header.sender_name == "position_estimate"){
					ProcessSelfLocationData(recv);
					is_update = true;
				}else if(header.sender_name == "control_motor"){
					ProcessSelfLocationData(recv);
					is_update = true;
				}else{
					cerr << "unknown data type received. \"" << header.type << "\"" << endl;
				}
			}else{
				cerr << "Received data is NOT exist in receive list (" << header.id << ")." << endl;
			}
		}
	}
	if (is_update){
		SendResult();
	}

	return RTC::RTC_OK;
}

RTC::ReturnCode_t EnvRecognition::
ProcessURGData(const std::string &recv)
{
	int ret;

	ret = SensorData->DeserializeFromXML(recv);
	if(ret != 0){ // データ受信失敗
		cerr << "URGData recv failed (" << ret << ")." << endl;
		return RTC::RTC_ERROR;
	}

	fusion.setGridMap(&SensorData->Data);

	if(Config->gui){
		set_value(&Result->Data);
	}

	return RTC::RTC_OK;
}

RTC::ReturnCode_t EnvRecognition::
ProcessSelfLocationData(const std::string &recv)
{
	int ret;

#if 0
	cerr << "PosData process." << endl;
	cerr << recv << endl;
#endif

	ret = PosData->DeserializeFromXML(recv);
	if(ret != 0){
		return RTC::RTC_ERROR;
	}

#if 0
	cout << "----- received PosData -----" << endl;
	cout << PosData->Data;
	cout << "------------------------------" << endl;
#endif

	fusion.setSelfLocalization(&PosData->Data);

	return RTC::RTC_OK;
}

RTC::ReturnCode_t EnvRecognition::
SendResult(void)
{
	struct timeval tv;
	string xml;

	gettimeofday(&tv,NULL);
	m_out.tm.sec = tv.tv_sec;
	m_out.tm.nsec = tv.tv_usec*1000;

	fusion.getMergedGridMap(&Result->Data);

	Result->SerializeToXML(&xml);

#if 0
	cout << "===== XML to send. =====" << endl;
	cout << xml;
	cout << "========================" << endl;
#endif

	m_out.data = xml.c_str();
	m_outOut.write();

	return RTC::RTC_OK;
}

extern "C"
{
 
	void EnvRecognitionInit(RTC::Manager* manager)
	{
		RTC::Properties profile(envrecognition_spec);
		manager->registerFactory(profile,
		                         RTC::Create<EnvRecognition>,
		                         RTC::Delete<EnvRecognition>);
	}
  
};

