// -*- C++ -*-
/*!
 * @file  LocalPathPlan.cpp * @brief LocalPathPlan * $Date$ 
 *
 * $Id$ 
 */
#include "LocalPathPlan.h"
#include "LocalPathPlanDialog.h"

#include "baseRTC.h"
#include "SerDesIF.h"
#include "Potential.h"

#include <iostream>

const char LocalPathPlan::Category[] = "navigation";
const char LocalPathPlan::Name[]     = "local_path_plan";

using namespace std;

// Module specification
// <rtc-template block="module_spec">
static const char* localpathplan_spec[] =
  {
    "implementation_id", "LocalPathPlan",
    "type_name",         "LocalPathPlan",
    "description",       "LocalPathPlan",
    "version",           "1.0.0",
    "vendor",            "Chiba Institute of Technology",
    "category",          "navigation",
    "activity_type",     "SPORADIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.filename", "./LocalPathPlan.xml",
    ""
  };
// </rtc-template>

LocalPathPlan::
LocalPathPlan(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    //m_inIn("in", m_in),
    m_outOut("out", m_out)

    // </rtc-template>
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  // registerInPort("in", m_inIn);

  // Set OutPort buffer
  registerOutPort("out", m_outOut);

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>
}

LocalPathPlan::
~LocalPathPlan()
{
	vector<RTC::TimedString *>::iterator it_m_in;
	vector<RTC::InPort<RTC::TimedString> *>::iterator it_m_inIn;

	if(Config != NULL){
		delete Config;
	}

	if(ResultPath != NULL){
		delete ResultPath;
	}

	if(PotentialMap != NULL){
		delete PotentialMap;
	}

	if(RiskMap != NULL){
		delete RiskMap;
	}

	if(PathData != NULL){
		delete PathData;
	}

	it_m_in = m_in.begin();
	while(it_m_in != m_in.end()){
		delete *it_m_in;
		it_m_in++;
	}

	it_m_inIn = m_inIn.begin();
	while(it_m_inIn != m_inIn.end()){
		delete *it_m_inIn;
		it_m_inIn++;
	}
}


RTC::ReturnCode_t LocalPathPlan::
onInitialize()
{
	int i;
	int ret;

	bindParameter( "filename", filename, "./LocalPathPlan.xml");

	Config = new LocalPathPlanConfig();

	ret = Config->LoadFromFile(filename);
	if(ret < 0){
		cerr << "can't load config file \"" << filename << "\"" << endl;
		delete Config;
		Config = NULL;
		return RTC::RTC_ERROR;
	}

	for(i=0; i<Config->input_number; i++){
		char buf[10];
		m_in.push_back(new RTC::TimedString);
		sprintf(buf, "in%d", i);
		m_inIn.push_back(new RTC::InPort<RTC::TimedString>(buf, *m_in[i]));
		registerInPort(buf, *m_inIn[i]);
	}

	delete Config;
	Config = NULL;
	
	return RTC::RTC_OK;
}

pthread_t tid;

RTC::ReturnCode_t LocalPathPlan::
onActivated(RTC::UniqueId ec_id)
{
	int ret;

	Config = new LocalPathPlanConfig();

	ret = Config->LoadFromFile(filename);
	if(ret < 0){
		cerr << "can't load config file \"" << filename << "\"" << endl;
		delete Config;
		Config = NULL;
		return RTC::RTC_ERROR;
	}

	// 設定ファイル内のカテゴリ名とコンポーネント名をチェック
	if(Config->category != LocalPathPlan::Category){
		cerr << "Category NOT match." << endl;
		delete Config;
		Config = NULL;
		return RTC::RTC_ERROR;
	}
	if(Config->name != LocalPathPlan::Name){
		cerr << "Name NOT match." << endl;
		delete Config;
		Config = NULL;
		return RTC::RTC_ERROR;
	}

	// 経路計画初期化
	Potential.init();
	Potential.setConfiguration(Config);

	// GUI初期化
	if (Config->gui){
		View = Config->grid_map; // 送信データサイズ設定
		View.map.clear();
		View.resize();

		initialize_window();
		pthread_create(&tid,NULL,windowThread,NULL);
	}

	// データ送信用インタフェイスクラス初期化
	ResultPath = new SerDesIF<Path>(LocalPathPlan::Category, LocalPathPlan::Name, Config->id);
	PotentialMap = new SerDesIF<GridMap>(LocalPathPlan::Category, LocalPathPlan::Name, Config->id);

	PotentialMap->Data = Config->grid_map;
	PotentialMap->Data.map.clear();
	PotentialMap->Data.resize();

	// データ受信用インタフェイスクラス初期化
	RiskMap  = new SerDesIF<GridMap>();
	PathData = new SerDesIF<Path>();
	PosData = new SerDesIF<Localization>();

	return RTC::RTC_OK;
}

RTC::ReturnCode_t LocalPathPlan::
onDeactivated(RTC::UniqueId ec_id)
{
	// データ送信用インタフェイスクラス初期化
	if(ResultPath != NULL){
		delete ResultPath;
		ResultPath = NULL;
	}

	if(PotentialMap != NULL){
		delete PotentialMap;
		PotentialMap = NULL;
	}

	// データ受信用インタフェイスクラス終了処理
	if(RiskMap != NULL){
		delete RiskMap;
		RiskMap = NULL;
	}
	if(PathData != NULL){
		delete PathData;
		PathData = NULL;
	}

	// GUI終了処理
	if (Config->gui){
		close_window();
		pthread_join(tid, NULL);
	}

	// 経路計画終了処理
	Potential.close();

	// コンフィグ破棄
	if(Config != NULL){
		delete Config;
		Config = NULL;
	}

	return RTC::RTC_OK;
}

RTC::ReturnCode_t LocalPathPlan::
onExecute(RTC::UniqueId ec_id)
{
	baseRTC::Header header;
	string recv;

	struct timeval tv;
	static double last_time;
	double time;
	gettimeofday(&tv,NULL);
	time = tv.tv_sec + tv.tv_usec / 1000000.0f;

	int i;
	for(i = 0; i < Config->input_number; i++){
		if (m_inIn[i]->isNew()){
			m_inIn[i]->read();
			recv = m_in[i]->data;

#if 0
			cout << "+++++ XML to recv. +++++" << endl;
			cerr << recv << endl;
			cout << "++++++++++++++++++++++++" << endl;
#endif

			header.readFromXml(recv);

			// 受信したデータのIDが, 設定ファイルのreceive 項目にあれば処理
			if(find(Config->receive.begin(), Config->receive.end(), header.id) != Config->receive.end()){
				// ヘッダ内のデータ情報から受信データの型を取得
				if(header.sender_name == "control_risk"){
					ProcessRiskMapData(recv);
				}else if(header.sender_name == "GlobalNavigationRTC"){
					if (header.type == Path::TYPE_ID) {
						cout << Path::TYPE_ID << " from "
							 << header.sender_name << "!\n";
						ProcessPathNodesData(recv);
					} else {
						cout << header.type 
							 << " from GlobalNavigationRTC ignored\n";
					}
				}else if(header.type == Localization::TYPE_ID){
					ProcessSelfLocationData(recv);
				}else{
					cerr << "unknown data type received. \"" << header.type << "\"" << endl;
				}
			}else{
				cerr << "Received data is NOT exist in receive list (" << header.id << "::" << header.sender_name << ")." << endl;
			}
			last_time = time;
		}
	}

	if ((time - last_time) > Config->timeout){
		Path path;
		path.path.resize(1);
		path.path[0].vel = Point(0,0,0);
		ResultPath->Data = path;
		last_time = time;
	}

	return RTC::RTC_OK;
}

RTC::ReturnCode_t LocalPathPlan::
ProcessRiskMapData(const std::string &recv)
{
	int ret;

	//int in_x, in_y;

	ret = RiskMap->DeserializeFromXML(recv);
	if(ret != 0){ // 復元失敗
		cerr << "Data deserialize failed (RiskMapData)[" << ret << "]." << endl;
		return RTC::RTC_ERROR;
	}

	Potential.setGridMap(&RiskMap->Data);

	return RTC::RTC_OK;
}

RTC::ReturnCode_t LocalPathPlan::
ProcessSelfLocationData(const std::string &recv)
{
	int ret;

#if 0
	cerr << "PosData process." << endl;
	cerr << recv << endl;
#endif

	ret = PosData->DeserializeFromXML(recv);
	if(ret != 0){
		return RTC::RTC_ERROR;
	}

#if 0
	cout << "----- received PosData -----" << endl;
	cout << PosData->Data;
	cout << "------------------------------" << endl;
#endif

	Potential.setLocalization(&PosData->Data);

	Potential.getLocalPath(&ResultPath->Data);

	// 送信データの時刻設定 (受信データの時刻に合わせる)
	ResultPath->SetTime(RiskMap->Data.time);

//	cerr << "SendResult()" << endl;
	SendResult();

	if(Config->gui){
		Potential.getPotentialMap(&View);
		set_value(&View);         // グリッドマップ
		set_value(&ResultPath->Data); // 経路
	}

	return RTC::RTC_OK;
}


RTC::ReturnCode_t LocalPathPlan::
SendResult(void)
{
	struct timeval tv;
	string xml;

	gettimeofday(&tv,NULL);
	m_out.tm.sec = tv.tv_sec;
	m_out.tm.nsec = tv.tv_usec*1000;

	ResultPath->SerializeToXML(&xml);

#if 0
	cout << "===== XML to send. =====" << endl;
	cout << xml;
	cout << "========================" << endl;
#endif

	m_out.data = xml.c_str();
	m_outOut.write();

	return RTC::RTC_OK;
}

RTC::ReturnCode_t LocalPathPlan::
ProcessPathNodesData(const std::string &recv)
{
	int ret;

	ret = PathData->DeserializeFromXML(recv);
	if(ret != 0){ // 復元失敗
		cerr << "Data deserialize failed (LocalPathPlan)[" << ret << "]." << endl;
		return RTC::RTC_ERROR;
	}

	std::cout << "******************* Get Global Path ******************" << std::endl;
	std::cout << PathData->Data;

	Potential.setGlobalPath(&PathData->Data);

	return RTC::RTC_OK;
}

extern "C"
{
 
  void LocalPathPlanInit(RTC::Manager* manager)
  {
    RTC::Properties profile(localpathplan_spec);
    manager->registerFactory(profile,
                             RTC::Create<LocalPathPlan>,
                             RTC::Delete<LocalPathPlan>);
  }
  
};
