// Potential.cpp : 実装ファイル
//
#include <sys/time.h>
#include <math.h>

#include "Potential.h"
#include "LocalPathPlanConfig.h"

#define len(x,y) sqrt((x)*(x)+(y)*(y))
#define map(x, y, nx, ny) map[(ny) * ((nx) - (x) - 1) + ((ny) - (y) - 1)]
#define PI 3.14159f

using namespace std;

// Potential
CPotential::CPotential():
	update(true)
{
}

CPotential::~CPotential()
{
}

// Potential メッセージ ハンドラ

// ■ 初期化 ■■■■■
int CPotential::
init()
{
	local_path.path.resize(1);
	dist_path.path.push_back(Node(0,0,0));

	return 0;
}

// ■ 設定 ■■■■■
int CPotential::
setConfiguration(LocalPathPlanConfig *conf)
{
	slope = conf->slope;
	goal_radius = conf->goal_radius;
	potential_map = conf->grid_map;
	potential_map.map.clear();
	potential_map.resize();

	return 0;
}

// ■ 入力 ■■■■■
// 衝突リスクを加味した障害物グリッドマップ
int CPotential::
setGridMap(GridMap *gmap)
{
	grid_map = *gmap;

	update = false;	// updatenが必要

	return 0;
}

// 大経路入力
int CPotential::
setGlobalPath(Path *gpath)
{
	global_path = *gpath;
	dist_path = *gpath;

	update = false;	// updateが必要

	return 0;
}

// 現在の位置の入力
int CPotential::setLocalization(Localization *localization)
{
	loc = *localization;

	update = false;	// updatenが必要

	return 0;
}

// ■ 出力 ■■■■■
// 経路データ取得
int CPotential::
getLocalPath(Path *lpath)
{
	if (!update){ // 未update -> update
		calculatePotentialMap(); // ポテンシャル地図の更新
		calculatePath();
		update = true; // update済
	}

	*lpath = local_path;

	return 0;
}

// ポテンシャルグリッドマップの取得
int CPotential::
getPotentialMap(GridMap *pmap)
{
	if (!update){ // 未update -> update
		calculatePotentialMap(); // ポテンシャル地図の更新
		calculatePath();
		update = true; // update済
	}

	*pmap = potential_map;

	return 0;
}

// ■ 終了処理 ■■■■■

int CPotential::
close()
{
	return 0;
}

// ■ private ■■■■■
//	ポテンシャル地図の更新
int CPotential::
calculatePotentialMap()
{
	if(grid_map.map.empty()){ // 地図入力がない場合
		return -1;
	}
	if(potential_map.map.empty()){ // 地図入力がない場合
		return -1;
	}

	GridMap work_map = potential_map;
	work_map.map.clear();
	work_map.resize();

	work_map.Mapping(grid_map);
	work_map.FillRectangle(-0.65f, -0.45f, 0.65f, 0.45f, 0.0f);
//	work_map.FillRectangle(-1.0f, -1.0f, 1.0f, 0.40f, 0.0f);

	int x, y, frame_x, frame_y, ix, iy, fx, fy, mx, my, i;
	int nx, ny;
	float p, res;
	vector<float> val;

	frame_x = (int)(1.0f / slope / work_map.size_x);	// ポテンシャルの傾きから求められる距離 
	frame_y = (int)(1.0f / slope / work_map.size_y);

	val.resize(frame_x * frame_y);
	for(y = 0; y < frame_y;y ++){						// 予め計算したポテンシャルを保存するための領域
		for(x = 0;x < frame_x;x ++){
			val[frame_x * y + x] = min(1.0f*len(x * work_map.size_x, y * work_map.size_y) * slope, 1.0f);
		}												// 計算して入れておく
	}

	nx = work_map.sizeX();
	ny = work_map.sizeY();

	for(i = 0; i < nx * ny; i ++)
		potential_map.map[i] = 0.0f;

	for(y = 0; y < ny; y ++){							// y方向，x方向にずらしながら，値を重ね合わせる
		for(x = 0; x < nx; x ++){
			res = work_map.map[nx * y + x];
			fx = frame_x * res / 1.0f;					// 重ね合わせに必要なフレームを計算
			fy = frame_y * res / 1.0f;
			for(iy = -fy+1; iy < fy;iy ++){
				for(ix = -fx+1;ix < fx;ix ++){
					mx = x + ix;
					my = y + iy;
					if ((mx < 0)||(mx >= nx)||(my < 0)||(my >= ny)) continue;
					p = res - val[frame_x * abs(iy) + abs(ix)];
					potential_map.map[nx * my + mx] = max(p, potential_map.map[nx * my + mx]);
				}										// 最大値で重ね合わせる．
			}
		}
	}

	return 0;
}

//	ポテンシャル地図の更新
int CPotential::calculatePath()
{
	int x, y, i, nx, ny;
	vector<float> val;

	static float old_x = 0, old_y = 0, prev_x = 0, prev_y = 0;
	float px, py, dx = 0, dy = 0, dt = 0, NUM_DIR = 16.0f;
	float check_point[5][2];
	int chk_x, chk_y;

	nx = potential_map.sizeX();
	ny = potential_map.sizeY();

	// x                      y
	check_point[0][0] = 0.70f; check_point[0][1] =  0.50f;
	check_point[1][0] = 0.70f; check_point[1][1] = -0.50f;
	check_point[2][0] = 0.70f; check_point[2][1] =  0.0f;
	check_point[3][0] = 0.0f; check_point[3][1] =  0.50f;
	check_point[4][0] = 0.0f; check_point[4][1] = -0.50f;
	px = py = 0;

	for(i = 0;i < 5;i ++){
		chk_x = ( check_point[i][0] - potential_map.area_x0)/ potential_map.size_x;	// 画素のx座標
		chk_y = ( check_point[i][1] - potential_map.area_y0)/ potential_map.size_y;	// 画素のy座標
		float cen_p = potential_map.map(chk_x, chk_y, nx, ny);
		if (cen_p <= 0.2) continue;

//		float ang, min_ang = 0.0f, sl, min_sl = 0.0, l, search_r = 1.0f;
//		for(ang = 0.0f; ang < 2.0f*M_PI; ang += 2.0f*M_PI/NUM_DIR){
//			sl = 0.0f;
//			for(l = search_r/10.0f;l <= search_r;l += search_r/10.0f){
//				x = (int)(l * cos(ang) / potential_map.size_x);
//				y = (int)(l * sin(ang) / potential_map.size_y);
//				sl += (potential_map.map(x + chk_x, y + chk_y, nx, ny) - cen_p) / l;
//			}
//			sl /= 10.0f;
//			if (sl < min_sl){
//				min_sl = sl;
//				min_ang = ang;
//			}
//		}
		float a = atan2(check_point[i][1], check_point[i][0]);
		px += cen_p * cos(a+PI);
		py += cen_p * sin(a+PI);
	}

//	for(i = 0;i < 5;i ++){
//		potential_map.DrawPoint(check_point[i][0], check_point[i][1], 1.0f);
//	}

	px = min(max(px*0.5f, -1.0f), 1.0f);
	py = min(max(py*0.5f, -1.0f), 1.0f);

//	potential_map.DrawLine(0, 0, px, py, 1.0f);

	Point dist;
	if (dist_path.path.size() > 0){
		float spd, max_spd = 0.6;
		dist = dist_path.path[0].node;
		dx = dist.x - loc.pos_x;
		dy = dist.y - loc.pos_y;
		if ((dist.x != prev_x)||(dist.y != prev_y)){
			old_x = prev_x;
			old_y = prev_y;
		}
		dt = atan2(dist.y - old_y, dist.x - old_x) - loc.pos_the;
		prev_x = dist.x;
		prev_y = dist.y;
		spd = sqrt(dx * dx + dy * dy);
		if (spd < goal_radius){
			dist_path.path.erase(dist_path.path.begin());
			dx = dy = 0.0f;
		}
		if (spd > max_spd){
			dx *= max_spd/spd;
			dy *= max_spd/spd;
		}
	}
	px +=   dx * cos(loc.pos_the) + dy * sin(loc.pos_the);
	py += - dx * sin(loc.pos_the) + dy * cos(loc.pos_the);
	dx = px;
	dy = py;

	while(dt >  PI) dt -= 2.0f*PI;
	while(dt < -PI) dt += 2.0f*PI;
	dt = min(max(dt * 2.0f, -1.0f), 1.0f);

	local_path.path[0].vel.x = dx;
	local_path.path[0].vel.y = dy;
	local_path.path[0].vel.t = dt;

	return 0;
}
